import time
import traceback

from os import walk
from os.path import getmtime, join, exists

import schedule

from lifeguard.logger import lifeguard_logger as logger
from lifeguard.validations import VALIDATIONS, load_validations, clear_validations
from lifeguard.settings import LIFEGUARD_DIRECTORY

VALID_TIME_PERIODS = [
    "seconds",
    "minutes",
    "hours",
    "days",
    "weeks",
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
]

MOMENTS = [
    "second",
    "minute",
    "hour",
    "day",
    "week",
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
]

WATCHED_FILES = {}
FOREVER = True


def configure_validations():
    for validation in VALIDATIONS:
        content = VALIDATIONS[validation]
        if "every" in content["schedule"]:
            time_period = get_time_period(content)
            if time_period in VALID_TIME_PERIODS:
                job_instance = schedule.every(content["schedule"]["every"][time_period])
                job_func = get_dynamic_job_func(job_instance, time_period)
                job_func.do(content["ref"]).tag("validation")
        if "at" in content["schedule"]:
            time_moment = get_time_moment(content)
            if time_moment in MOMENTS:
                moment = getattr(schedule.every(), time_moment)
                moment.at(content["schedule"]["at"][time_moment]).do(
                    content["ref"]
                ).tag("validation")


def get_dynamic_job_func(job_instance, time_period):
    return getattr(job_instance, time_period)


def get_time_period(content):
    keys = content["schedule"]["every"].keys()
    time_period = list(keys)[0]
    return time_period


def get_time_moment(content):
    keys = content["schedule"]["at"].keys()
    time_moment = list(keys)[0]
    return time_moment


def __load_validations_files():
    for root, _dirs, files in walk(join(LIFEGUARD_DIRECTORY, "validations")):
        for file in files:
            if file.endswith(".yaml"):
                file_path = join(root, file)
                if file_path not in WATCHED_FILES:
                    logger.info("watching file %s", file_path)
                    WATCHED_FILES[file_path] = getmtime(file_path)


def check_if_should_reload():
    changed = False
    # check new files or changed files
    for root, _dirs, files in walk(join(LIFEGUARD_DIRECTORY, "validations")):
        for file in files:
            if file.endswith(".yaml"):
                file_path = join(root, file)
                if file_path not in WATCHED_FILES:
                    logger.info("watching file %s", file_path)
                    WATCHED_FILES[file_path] = getmtime(file_path)
                    changed = True
                else:
                    last_modified = getmtime(file_path)
                    if last_modified != WATCHED_FILES[file_path]:
                        logger.info("file changed %s", file_path)
                        WATCHED_FILES[file_path] = last_modified
                        changed = True
    # check removed files
    for file_path in list(WATCHED_FILES.keys()):
        if not exists(file_path):
            logger.info("file removed %s", file_path)
            del WATCHED_FILES[file_path]
            changed = True

    if changed:
        reload_scheduler()


def __prepend_reload_job():
    schedule.every(15).seconds.do(check_if_should_reload).tag("lifeguard")


def __clear_validations_jobs():
    schedule.clear("validation")


def reload_scheduler():
    __clear_validations_jobs()
    clear_validations()
    load_validations()
    configure_validations()


def start_scheduler():
    __load_validations_files()
    __prepend_reload_job()

    configure_validations()

    while FOREVER:
        time.sleep(1)
        try:
            schedule.run_pending()
        except Exception as exception:
            logger.error(
                "error on execute scheduler %s",
                str(exception),
                extra={"traceback": traceback.format_exc()},
            )



"""
Lifeguard Context Implementation
"""


class LifeguardContext:
    """
    Lifeguard Context
    """

    def __init__(self):
        self._users = []
        self._auth_method = None
        self._only_settings = False
        self._cors_settings = None
        self._alert_email_template = """From: [[SMTP_USER]]
To: [[RECEIVERS]]
MIME-Version: 1.0
Content-type: text/html
Subject: [[SUBJECT]]

<h1>Validation details</h1>

<quote>
[[MESSAGE]]
</quote>

<hr/>
<p>
    See more details in Lifeguard Dashboard: [[LIFEGUARD_PUBLIC_ADDRESS]].
</p>
"""

    @property
    def only_settings(self):
        """
        Getter for only settings
        """
        return self._only_settings

    @only_settings.setter
    def only_settings(self, value):
        """
        Setter for only settings
        """
        self._only_settings = value

    @property
    def alert_email_template(self):
        """
        Getter for email template
        """
        return self._alert_email_template

    @alert_email_template.setter
    def alert_email_template(self, value):
        """
        Setter for alert email template
        """
        self._alert_email_template = value

    @property
    def auth_method(self):
        """
        Getter for auth method
        """
        return self._auth_method

    @auth_method.setter
    def auth_method(self, value):
        """
        Setter for auth method
        """
        self._auth_method = value

    @property
    def users(self):
        """
        Getter for users
        """
        return self._users

    @users.setter
    def users(self, value):
        """
        Setter for users
        """
        self._users = value

    @property
    def cors_settings(self):
        """
        Getter for cors settings
        """
        return self._cors_settings

    @cors_settings.setter
    def cors_settings(self, value):
        """
        Setter for cors settings
        """
        self._cors_settings = value

    def valid_user(self, username, password):
        """
        Check if user is valid
        """
        return [
            user
            for user in self.users
            if user["username"] == username and user["password"] == password
        ]


LIFEGUARD_CONTEXT = LifeguardContext()



import unittest
from unittest.mock import patch


from lifeguard import (
    change_status,
    setup,
    NORMAL,
    WARNING,
    PROBLEM,
    ACTION_STATUSES,
)
from lifeguard.context import LIFEGUARD_CONTEXT

from tests.fixtures import mock_lifeguard_settings


class TestLifeguardCore(unittest.TestCase):
    def test_statuses(self):
        self.assertEqual(NORMAL, "NORMAL")
        self.assertEqual(WARNING, "WARNING")
        self.assertEqual(PROBLEM, "PROBLEM")
        self.assertEqual(ACTION_STATUSES, [NORMAL, WARNING, PROBLEM])

    def test_change_status(self):
        self.assertEqual(change_status(NORMAL, NORMAL), NORMAL)
        self.assertEqual(change_status(NORMAL, WARNING), WARNING)
        self.assertEqual(change_status(PROBLEM, WARNING), PROBLEM)

    @patch("lifeguard.load_validations")
    @patch("lifeguard.recover_settings")
    def test_setup_call_load_validations(
        self, mock_recover_settings, mock_load_validations
    ):
        mock_recover_settings.return_value = mock_lifeguard_settings

        setup(LIFEGUARD_CONTEXT)
        mock_load_validations.assert_called()



import unittest
from unittest.mock import patch, MagicMock, call

from lifeguard.scheduler import (
    configure_validations,
    VALID_TIME_PERIODS,
    MOMENTS,
    check_if_should_reload,
    reload_scheduler,
    start_scheduler,
    WATCHED_FILES,
)

mock_ref = MagicMock(name="ref")


class TestScheduler(unittest.TestCase):
    @patch(
        "lifeguard.scheduler.VALIDATIONS",
        {"example": {"ref": mock_ref, "schedule": {"every": {"minutes": 1}}}},
    )
    @patch("lifeguard.scheduler.schedule")
    def test_configure_validations_every_minute(self, mock_schedule):
        configure_validations()

        mock_schedule.every.assert_called_with(1)

    @patch(
        "lifeguard.scheduler.VALIDATIONS",
        {"example": {"ref": mock_ref, "schedule": {"every": {"hours": 1}}}},
    )
    @patch("lifeguard.scheduler.schedule")
    def test_configure_validations_every_hour(self, mock_schedule):
        configure_validations()

        mock_schedule.every.assert_called_with(1)

    @patch(
        "lifeguard.scheduler.VALIDATIONS",
        {"example": {"ref": mock_ref, "schedule": {"every": {"blah": 1}}}},
    )
    @patch("lifeguard.scheduler.schedule")
    def test_configure_validations_with_invalid_time_period(self, mock_schedule):
        configure_validations()

        mock_schedule.every.assert_not_called()

    @patch(
        "lifeguard.scheduler.VALIDATIONS",
        {"example": {"ref": mock_ref, "schedule": {"at": {"hour": "10:00"}}}},
    )
    @patch("lifeguard.scheduler.schedule")
    def test_configure_validations_at_specific_minute(self, mock_schedule):
        mock_do = MagicMock(name="do")

        mock_at = MagicMock(name="at")
        mock_at.at.return_value = mock_do

        mock_return_of_every = MagicMock(name="mock_return_of_every")
        mock_schedule.every.return_value = mock_return_of_every
        mock_return_of_every.hour = mock_at

        configure_validations()

        mock_schedule.every.assert_called()
        mock_at.at.assert_called_with("10:00")
        mock_do.do.assert_called_with(mock_ref)

    def test_valid_time_periods(self):
        self.assertEqual(
            VALID_TIME_PERIODS,
            [
                "seconds",
                "minutes",
                "hours",
                "days",
                "weeks",
                "monday",
                "tuesday",
                "wednesday",
                "thursday",
                "friday",
                "saturday",
                "sunday",
            ],
        )

    def test_moments(self):
        self.assertEqual(
            MOMENTS,
            [
                "second",
                "minute",
                "hour",
                "day",
                "week",
                "monday",
                "tuesday",
                "wednesday",
                "thursday",
                "friday",
                "saturday",
                "sunday",
            ],
        )

    @patch("lifeguard.scheduler.clear_validations")
    @patch("lifeguard.scheduler.load_validations")
    @patch("lifeguard.scheduler.configure_validations")
    @patch("lifeguard.scheduler.schedule")
    def test_reload_scheduler(
        self,
        mock_schedule,
        mock_configure_validations,
        mock_load_validations,
        mock_clear_validations,
    ):
        reload_scheduler()

        mock_schedule.clear.assert_called_with("validation")
        mock_configure_validations.assert_called_with()
        mock_load_validations.assert_called_with()
        mock_clear_validations.assert_called_with()

    @patch("lifeguard.scheduler.FOREVER", False)
    @patch("lifeguard.scheduler.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.scheduler.configure_validations")
    @patch("lifeguard.scheduler.getmtime")
    def test_start_scheduler(self, mock_getmtime, mock_configure_validations):
        mock_getmtime.return_value = 1

        start_scheduler()

        mock_configure_validations.assert_called_with()
        self.assertEqual(
            WATCHED_FILES, {"tests/fixtures/validations/simple_validation.yaml": 1}
        )

    @patch("lifeguard.scheduler.WATCHED_FILES", {})
    @patch("lifeguard.scheduler.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.scheduler.reload_scheduler")
    @patch("lifeguard.scheduler.logger")
    def test_check_if_should_reload_with_new_file(
        self, mock_logger, mock_reload_scheduler
    ):
        check_if_should_reload()
        mock_reload_scheduler.assert_called_with()
        mock_logger.info.assert_has_calls(
            [
                call(
                    "watching file %s",
                    "tests/fixtures/validations/simple_validation.yaml",
                )
            ]
        )

    @patch(
        "lifeguard.scheduler.WATCHED_FILES",
        {"tests/fixtures/validations/simple_validation.yaml": 1},
    )
    @patch("lifeguard.scheduler.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.scheduler.reload_scheduler")
    @patch("lifeguard.scheduler.logger")
    def test_check_if_should_reload_with_file_changed(
        self, mock_logger, mock_reload_scheduler
    ):
        check_if_should_reload()
        mock_reload_scheduler.assert_called_with()
        mock_logger.info.assert_has_calls(
            [
                call(
                    "file changed %s",
                    "tests/fixtures/validations/simple_validation.yaml",
                )
            ]
        )

    @patch(
        "lifeguard.scheduler.WATCHED_FILES",
        {"tests/fixtures/validations/not_exists.yaml": 1},
    )
    @patch("lifeguard.scheduler.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.scheduler.reload_scheduler")
    @patch("lifeguard.scheduler.logger")
    def test_check_if_should_reload_with_file_removed(
        self, mock_logger, mock_reload_scheduler
    ):
        check_if_should_reload()
        mock_reload_scheduler.assert_called_with()
        mock_logger.info.assert_has_calls(
            [call("file removed %s", "tests/fixtures/validations/not_exists.yaml")]
        )



"""
Base of notification system
"""

from datetime import datetime

from lifeguard.logger import lifeguard_logger as logger

NOTIFICATION_METHODS = []

SINGLE_MESSAGE_NOTIFICATION = "single"
INIT_THREAD_MESSAGE_NOTIFICATION = "init_thread"
UPDATE_THREAD_MESSAGE_NOTIFICATION = "update_thread"
CLOSE_THREAD_MESSAGE_NOTIFICATION = "close_thread"
NOTIFICATION_TYPE = []


class NotificationOccurrence:
    """
    Notification history entry
    """

    def __init__(
        self, validation_name, details, status, notification_type, created_at=None
    ):
        self._validation_name = validation_name
        self._details = details
        self._status = status
        self._notification_type = notification_type
        self._created_at = created_at if created_at else datetime.now()

    @property
    def validation_name(self):
        """
        Return validation name
        """
        return self._validation_name

    @property
    def details(self):
        """
        Return details
        """
        return self._details

    @property
    def status(self):
        """
        Return status
        """
        return self._status

    @property
    def notification_type(self):
        """
        Return notification type
        """
        return self._notification_type

    @property
    def created_at(self):
        """
        Return created_at
        """
        return self._created_at


class NotificationStatus:
    """
    Notification status class
    """

    def __init__(self, validation_name, thread_ids, options=None, content=None):
        self._validation_name = validation_name
        self._thread_ids = thread_ids
        self._options = options
        self._opened = True
        self._last_notification = datetime.now()

    @property
    def validation_name(self):
        """
        Return validation name
        """
        return self._validation_name

    @property
    def thread_ids(self):
        """
        Return thread ids
        """
        return self._thread_ids

    @property
    def is_opened(self):
        """
        Return opened
        """
        return self._opened

    @is_opened.setter
    def is_opened(self, value):
        self._opened = value

    def update(self):
        """
        Close notification
        """
        self._last_notification = datetime.now()

    def close(self):
        """
        Close notification
        """
        self._last_notification = datetime.now()
        self._opened = False

    @property
    def options(self):
        """
        Return options
        """
        return self._options

    @property
    def last_notification(self):
        """
        Return last notification
        """
        return self._last_notification

    @last_notification.setter
    def last_notification(self, value):
        """
        Setter for last notification
        """
        self._last_notification = value

    def get_attributes(self):
        """
        Return all attributes in a dict.
        """
        attributes = {
            "validation_name": self.validation_name,
            "thread_ids": self.thread_ids,
            "options": self.options,
            "is_opened": self.is_opened,
        }

        if self.last_notification:
            attributes["last_notification"] = self.last_notification.strftime(
                "%Y-%m-%dT%H:%M"
            )
        return attributes

    def __str__(self):
        return str(self.get_attributes())


class NotificationBase:
    """
    Base of notification
    """

    @property
    def name(self):
        """
        Implementation name
        """
        return "not implemented"

    def send_single_message(self, content, settings):
        """
        Notify in single message
        """

    def init_thread(self, content, settings):
        """
        Notify in single message

        :return: str
        """

    def update_thread(self, thread_id, content, settings):
        """
        Update thread with a new message
        """

    def close_thread(self, thread_id, content, settings):
        """
        Close thread with a final message
        """


def append_notification_implementation(implementation):
    """
    Append a new implementation
    """
    if implementation.__name__ not in [
        notification_method.__class__.__name__
        for notification_method in NOTIFICATION_METHODS
    ]:
        logger.info("appending implementation %s", implementation.__name__)
        NOTIFICATION_METHODS.append(implementation())



import unittest
from unittest.mock import MagicMock, mock_open, patch

from lifeguard.bootstrap import generate_base_project


class TestBootstrap(unittest.TestCase):
    @patch("lifeguard.bootstrap.os")
    def test_generate_base_project(self, mock_os):
        mock_os.path.exists.return_value = False

        mocked_open = mock_open()

        with patch("builtins.open", mocked_open, create=True):
            generate_base_project()

        mocked_open.assert_called_with("lifeguard_settings.py", "w")

        mocked_file = mocked_open()
        mocked_file.write.assert_called_with(
            "\nPLUGINS = []\n\ndef setup(_lifeguard_context):\n    pass\n"
        )

        mock_os.path.exists.assert_called_with("validations")
        mock_os.makedirs.assert_called_with("validations")
